//
// SharedRep.java - Shared Repository
//

package nyangau.rep;

//...simports:0:
import java.lang.reflect.Constructor;
import java.util.*;
import java.util.regex.*;
import java.util.logging.*;
import java.io.*;
import java.net.*;
import java.security.*;
import java.security.cert.*;
import javax.net.*;
import javax.net.ssl.*;
import javax.security.auth.x500.*;
import java.lang.management.*;
import javax.management.*;

import javax.xml.XMLConstants;
import javax.xml.namespace.NamespaceContext;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.ErrorHandler;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.Text;

import nyangau.se.*;
//...e

public class SharedRep implements SharedRepMBean
  {
  private static String VERSION = "0.9";
  private static Logger logger = Logger.getLogger(SharedRep.class.getName());
  protected SocketFactory sf;
  protected ServerSocket ss;
  protected int loc;
  protected String password;
  protected String password2;
  protected long tAdjust;
  protected Object oSync;
  protected Rep rep;
  protected Map<String,MapValue> map;
  protected Map<Integer,SharedRepLocation> locations;
  protected static final byte FT_FILE      = 'F';
  protected static final byte FT_TOMBSTONE = 'T';
  protected static boolean caseSensitive;
  protected static boolean verifyHostname;
  protected java.util.concurrent.BlockingQueue<FetchRequest> fetchq;
  protected boolean fetching;
  protected SharedRepTrigger trigger;
  protected int triggerCount;
  protected long triggerTime;
//...sgetLoc:2:
// Part of SharedRepMBean interface

public int getLoc()
  {
  return loc;
  }
//...e
//...sgetPassword:2:
public String getPassword()
  {
  return password;
  }
//...e
//...sgetTimeAdjustment:2:
public long getTimeAdjustment()
  { return tAdjust; }
//...e
//...sgetMapKeys:2:
// Add all the entries in the map to the set
public Set<String> getMapKeys()
  {
  synchronized ( oSync )
    {
    return map.keySet();
    }
  }
//...e
//...sgetMapValue:2:
public MapValue getMapValue(String name)
  {
  synchronized ( oSync )
    {
    return map.get(name);
    }
  }
//...e
//...sconnectTo:2:
//...sgetCN:2:
protected static String getCN(String name)
  throws IOException
  {
  StringTokenizer st = new StringTokenizer(name, ",");
  while ( st.hasMoreTokens() )
    {
    String s = st.nextToken();
    int i = s.indexOf("CN=");
    if ( i >= 0 )
      return s.substring(i+3);
    }
  throw new IOException("DN has no CN");
  }
//...e
//...scheckCertificate:2:
protected static void checkCertificate(String dns, java.security.cert.X509Certificate cert)
  throws IOException
  {
  String cn = getCN( cert.getSubjectDN().getName() );
  if ( dns.equals(cn) )
    return;
  try
    {
    Collection c = cert.getSubjectAlternativeNames();
    if ( c != null )
      {
      Iterator i = c.iterator();
      while ( i.hasNext() )
        {
        List l = (List) i.next();
        if ( ((Integer)l.get(0)).intValue() == 2 &&
             dns.equals((String)l.get(1)) )
          return;
        }
      }
    throw new IOException("DNS does not match CN in certificate DN or alternate subject DNS names");
    }
  catch ( CertificateParsingException e )
    {
    throw new IOException("Error parsing certificate");
    }
  }
//...e
//...scheckSession:2:
protected static void checkSession(SSLSession ssess)
  throws IOException
  {
  try
    {
    java.security.cert.Certificate[] certs = ssess.getPeerCertificates();
    if ( certs == null || certs.length == 0 )
      throw new IOException("No certificates");
    if ( ! ( certs[0] instanceof java.security.cert.X509Certificate ) )
      throw new IOException("Expected X509 certificate");
    String dns = ssess.getPeerHost();
    checkCertificate(dns, (java.security.cert.X509Certificate) certs[0]);
    }
  catch ( SSLPeerUnverifiedException e )
    {
    throw new IOException("Cannot establish trust in certificates");
    }
  }
//...e

public Socket connectTo(String host, int port)
  throws IOException
  {
  Socket s = sf.createSocket(host, port);
  if ( verifyHostname )
    {
    SSLSocket ss = (SSLSocket) s;
    SSLSession ssess = ss.getSession();
    checkSession(ssess);
    }
  return s;
  }
//...e
//...smarkDirty:2:
// Tell all other locations the latest value for this name in our map
public void markDirty(String d)
  {
  for ( Integer locI : locations.keySet() )
    locations.get(locI).markDirty(d);
  }
//...e
//...schangeMapValue:2:
public void changeMapValue(String name, MapValue mvalue)
  {
  synchronized ( oSync )
    {
    map.put(name, mvalue);
    markDirty(name);
    }
  }
//...e
//...smapValueChanged:2:
// The map for location locLocation has a new entry for name
public void mapValueChanged(int locLocation, int conn, String name)
  {
  SharedRepLocation l = locations.get(Integer.valueOf(locLocation));
  MapValue mvalue = l.getMapValue(name);
  synchronized ( oSync )
    {
    MapValue mvalueOld = map.get(name);
    if ( mvalueOld == null ||
         mvalueOld.compareTo(mvalue) < 0 )
      // We've not got it here
      // Or if we have, the other location is newer
      {
      if ( mvalue.ft == FT_TOMBSTONE )
        {
        logger.fine("SharedRep: mapValueChanged "+locLocation+" "+name+"->"+mvalue+", delete local");
        try
          {
          rep.deleteFile(name, mvalue.loc, mvalue.ts);
          changeMapValue(name, mvalue);
          if ( trigger != null )
            trigger.deleted(name);
          ++triggerCount;
          triggerTime = System.currentTimeMillis();
          }
        catch ( FileNotFoundException e )
          {
          }
        catch ( IOException e )
          {
          }
        }
      else if ( mvalue.ft == FT_FILE )
        {
        logger.fine("SharedRep: mapValueChanged "+locLocation+" "+name+"->"+mvalue+", queue fetch");
        try
          {
          fetchq.put(new FetchRequest(locLocation, conn, name, mvalue.loc, mvalue.ts));
          }
        catch ( InterruptedException e )
          // Won't happen, queue can grow forever
          {
          }
        }
      }
    }
  }
//...e

//...sclass FetchRequest:2:
public class FetchRequest
  {
  public int locRep;  // where to fetch from
  public int conn;    // connection number from that location
  public String name; // name of the file
  public int loc;
  public long ts;
  public FetchRequest(int lR, int c, String nF, int lF, long tF)
    { locRep = lR; conn = c; name = nF; loc = lF; ts = tF; }
  public String toString()
    { return "{locRep="+locRep+",conn="+conn+",name="+name+",loc="+loc+",ts="+ts+"}"; }
  }
//...e
//...sclass Fetcher:2:
public class Fetcher extends Thread
  {
//...sfetch:4:
// Returns bytes if fetched ok
// Returns null if couldn't get it, and no point retrying
// Throws IOException if error communicating with remote end

protected byte[] fetch(int locRep, String name, int loc, long ts)
  throws IOException
  {
  SharedRepLocation location = locations.get(Integer.valueOf(locRep));
  SharedRepLocation.Connection c = null;
  try
    {
    c = location.acquireConnection();
    c.dos.writeByte('r');
    c.dos.writeUTF(name);
    c.dos.writeInt(loc);
    c.dos.writeLong(ts);
    c.dos.flush();
    int rc = c.dis.readUnsignedByte();
    switch ( rc )
      {
      case 'N':
        location.releaseConnection(c);
        logger.fine("Fetcher: ("+name+","+loc+","+ts+") is no longer available at "+locRep);
        return null;
      case 'I':
        location.releaseConnection(c);
        logger.fine("Fetcher: ("+name+","+loc+","+ts+") not readable at "+locRep);
        return null;
      case 'r':
        {
        int n = c.dis.readInt();
        byte[] b = new byte[n];
        c.dis.readFully(b);
        location.releaseConnection(c);
        return b;
        }
      default:
        c.disconnect();
        logger.warning("Fetcher: ("+name+","+loc+","+ts+") from "+locRep+" unexpected response "+rc);
        return null;
      }
    }
  catch ( IOException e )
    {
    if ( c != null )
      try
        {
        c.disconnect();
        }
      catch ( IOException e2 )
        {
        }
    logger.fine("Fetcher: ("+name+","+loc+","+ts+") from "+locRep+" got IOException");
    }
  return null;
  }
//...e
//...srun:4:
public void run()
  {
  long backlogLast = -1;
  for ( ;; )
    try
      {
      fetching = false;
      FetchRequest fr = fetchq.take();
      fetching = true;
      SharedRepLocation l = locations.get(Integer.valueOf(fr.locRep));
      MapValue mvalue = l.getMapValue(fr.name);
      if ( mvalue != null && mvalue.ft == FT_FILE )
        // Remote file still on offer
        {
        MapValue mvalueOld;
        synchronized ( oSync )
          {
          mvalueOld = map.get(fr.name);
          }
        if ( mvalueOld == null ||
             mvalueOld.compareTo(mvalue) < 0 )
          // Our local copy (if we have one) is still out of date
          // (no update to local repository while request was queued)
          {
          try
            {
            logger.fine("Fetcher: fetching ("+fr.name+","+fr.loc+","+fr.ts+") from "+fr.locRep);
            long t0 = System.currentTimeMillis();
            byte[] b = fetch(fr.locRep, fr.name, fr.loc, fr.ts);
            long t1 = System.currentTimeMillis();
            if ( b != null )
              {
              synchronized ( oSync )
                {
                mvalueOld = map.get(fr.name);
                if ( mvalueOld == null ||
                     mvalueOld.compareTo(mvalue) < 0 )
                  // Our local copy (if we have one) is still out of date
                  // (no update to local repository while fetching)
                  {
                  rep.writeFile(fr.name, fr.loc, fr.ts, true, b);
                  long t2 = System.currentTimeMillis();
                  changeMapValue(fr.name, new MapValue(FT_FILE, fr.loc, fr.ts));
                  logger.fine("Fetcher: fetched "+b.length+" bytes in "+(t1-t0)+"ms and stored in "+(t2-t1)+"ms");
                  if ( trigger != null )
                    trigger.written(fr.name, b);
                  ++triggerCount;
                  triggerTime = System.currentTimeMillis();
                  }
                else
                  {
                  logger.fine("Fetcher: fetched "+b.length+" bytes in "+(t1-t0)+"ms, but local version changed");
                  }
                }
              }
            else
              logger.fine("Fetcher: fetch failed");
            }
          catch ( IOException e )
            {
            if ( fr.conn == l.getReceiverConn() )
              // Still have the same receiver connection upon which we were
              // notified of the change, so this could well be a transient
              // problem, and so we should retry.
              // We re-enqueue at the end of the queue so we process other
              // requests, hopefully to other locations first.
              {
              logger.fine("Fetcher: re-enqueue fetch");
              fetchq.put(fr);
              }
            }
          }
        else
          {
          logger.fine("Fetcher: local version updated whilst queued");
          }
        }
      else
        {
        logger.fine("Fetcher: remote file no longer on offer, not fetching "+fr);
        }
      int backlog = fetchq.size();
      if ( (backlog % 100) == 0 && backlog != backlogLast )
        {
        logger.info("Fetcher: backlog="+backlog);
        backlogLast = backlog;
        }
      }
    catch ( InterruptedException e )
      {
      }
  }
//...e
  }
//...e
//...sclass Connection:2:
// Accept and validate credentials
// Then accept commands until the connection is broken

public class Connection extends Thread
  {
  protected Socket s;
  public Connection(Socket s)
    {
    this.s = s;
    }
  public void run()
    {
    try
      {
      DataInputStream  dis = new DataInputStream (s.getInputStream ());
      DataOutputStream dos = new DataOutputStream(s.getOutputStream());
      String pw = dis.readUTF();
      if ( pw.equals(password) || pw.equals(password2) )
        ; // ok
      else
        {
        logger.warning("Connection: authentication error in inbound connection, connection terminated");
        dos.writeByte('S'); // security error
        dos.flush();
        s.close();
        return;
        }
      dos.writeByte(0); // login ok
      dos.flush();
      for ( ;; )
        {
        int cmd = dis.readUnsignedByte();
        switch ( cmd )
          {
          // from other SharedReps
//...sM \45\ metadata stream from other location:12:
case 'M':
  {
  int locC = dis.readInt();
  SharedRepLocation l = locations.get(Integer.valueOf(locC));
  if ( l != null )
    l.newReceiver(s, dis, dos);
  else
    {
    logger.warning("Connection: connection from location "+locC+", which we don't know about, connection terminated");
    s.close();
    }
  }
  return;
//...e
//...sr \45\ readFile with specific loc and ts:12:
case 'r':
  {
  String name = dis.readUTF();
  int loc = dis.readInt();
  long ts = dis.readLong();
  long t0 = System.currentTimeMillis();
  byte[] b;
  try
    {
    b = readFile(name, loc, ts);
    }
  catch ( FileNotFoundException e )
    {
    dos.writeByte('N');
    dos.flush();
    break;
    }
  catch ( IOException e )
    {
    dos.writeByte('I');
    dos.flush();
    break;
    }
  long t1 = System.currentTimeMillis();
  dos.writeByte('r');
  dos.writeInt(b.length);
  dos.write(b);
  dos.flush();
  long t2 = System.currentTimeMillis();
  logger.fine("Connection: read "+b.length+" bytes of ("+name+","+loc+","+ts+") in "+(t1-t0)+"ms and returned in "+(t2-t1)+"ms");
  }
  break;
//...e
          // from API client
//...sf \45\ isFile:12:
case 'f':
  {
  String name = dis.readUTF();
  dos.writeByte(isFile(name)?'f':'n');
  dos.flush();
  }
  break;
//...e
//...sd \45\ isDirectory:12:
case 'd':
  {
  String name = dis.readUTF();
  dos.writeByte(isDirectory(name)?'d':'n');
  dos.flush();
  }
  break;
//...e
//...sF \45\ filesInDirectory:12:
case 'F':
  {
  String name = dis.readUTF();
  String[] n;
  try
    {
    n = filesInDirectory(name);
    }
  catch ( FileNotFoundException e )
    {
    dos.writeByte('N');
    dos.flush();
    break;
    }
  catch ( IOException e )
    {
    dos.writeByte('I');
    dos.flush();
    break;
    }
  dos.writeByte('F');
  dos.writeInt(n.length);
  for ( int i = 0; i < n.length; i++ )
    dos.writeUTF(n[i]);
  dos.flush();
  }
  break;
//...e
//...sD \45\ directoriesInDirectory:12:
case 'D':
  {
  String name = dis.readUTF();
  String[] n;
  try
    {
    n = directoriesInDirectory(name);
    }
  catch ( FileNotFoundException e )
    {
    dos.writeByte('N');
    dos.flush();
    break;
    }
  catch ( IOException e )
    {
    dos.writeByte('I');
    dos.flush();
    break;
    }
  dos.writeByte('D');
  dos.writeInt(n.length);
  for ( int i = 0; i < n.length; i++ )
    dos.writeUTF(n[i]);
  dos.flush();
  }
  break;
//...e
//...sT \45\ fileModTs:12:
// Return the file modification time.
// This returns a local timestamp, not a cluster timestamp.

case 'T':
  {
  String name = dis.readUTF();
  long ts;
  try
    {
    ts = getTs(name) - getTimeAdjustment();
    }
  catch ( FileNotFoundException e )
    {
    dos.writeByte('N');
    dos.flush();
    break;
    }
  catch ( IOException e )
    {
    dos.writeByte('I');
    dos.flush();
    break;
    }
  dos.writeByte('T');
  dos.writeLong(ts);
  dos.flush();
  }
  break;
//...e
//...sR \45\ readFile:12:
case 'R':
  {
  String name = dis.readUTF();
  byte[] b;
  try
    {
    b = readFile(name);
    }
  catch ( FileNotFoundException e )
    {
    dos.writeByte('N');
    dos.flush();
    break;
    }
  catch ( IOException e )
    {
    dos.writeByte('I');
    dos.flush();
    break;
    }
  dos.writeByte('R');
  dos.writeInt(b.length);
  dos.write(b);
  dos.flush();
  }
  break;
//...e
//...sW \45\ writeFile:12:
case 'W':
  {
  String name = dis.readUTF();
  int l = dis.readInt();
  byte[] b = new byte[l];
  dis.readFully(b);
  try
    {
    writeFile(name, true, b);
    }
  catch ( FileNotFoundException e )
    {
    dos.writeByte('N');
    dos.flush();
    break;
    }
  catch ( IOException e )
    {
    dos.writeByte('I');
    dos.flush();
    break;
    }
  dos.writeByte('W');
  dos.flush();
  }
  break;
//...e
//...sZ \45\ deleteFile:12:
case 'Z':
  {
  String name = dis.readUTF();
  try
    {
    deleteFile(name);
    }
  catch ( FileNotFoundException e )
    {
    dos.writeByte('N');
    dos.flush();
    break;
    }
  catch ( IOException e )
    {
    dos.writeByte('I');
    dos.flush();
    break;
    }
  dos.writeByte('Z');
  dos.flush();
  }
  break;
//...e
//...sX \45\ findXPath:12:
case 'X':
  {
  String name       = dis.readUTF();
  String filter     = dis.readUTF();
  boolean recurse   = ( dis.readUnsignedByte() != 0 );
  boolean content   = ( dis.readUnsignedByte() != 0 );
  String xpath      = dis.readUTF();
  String nsprefixes = dis.readUTF();
  String xml;
  try
    {
    xml = findXPath(
      name,
      filter.equals("") ? null : filter,
      recurse,
      content,
      xpath,
      nsprefixes.equals("") ? null : nsprefixes
      );
    }
  catch ( FileNotFoundException e )
    {
    dos.writeByte('N');
    dos.flush();
    break;
    }
  catch ( IOException e )
    {
    dos.writeByte('I');
    dos.flush();
    break;
    }
  catch ( PatternSyntaxException e )
    {
    dos.writeByte('P');
    dos.flush();
    break;
    }
  catch ( XPathExpressionException e )
    {
    dos.writeByte('Q');
    dos.flush();
    break;
    }
  dos.writeByte('X');
  // We don't use writeUTF, as we can have 64KB or more bytes of data
  byte[] b = xml.getBytes("UTF-8");
  dos.writeInt(b.length);
  dos.write(b);
  dos.flush();
  }
  break;
//...e
//...sdefault:12:
default:
  logger.warning("Connection: unknown cmd="+cmd+", connection terminated");
  s.close();
  return;
//...e
          }
        }
      }
    catch ( IOException e )
      {
      try
        {
        s.close();
        }
      catch ( IOException e2 )
        {
        }
      }
    }
  }
//...e
//...sclass Server:2:
public class Server extends Thread
  {
  public void run() 
    {
    for ( ;; )
      try
        {
        Connection c = new Connection(ss.accept());
        c.start();
        }
      catch ( IOException e )
        {
        logger.fine("Server: IOException accepting connection");
        }
    }
  }
//...e
//...sclass WebServer:2:
public class WebServer extends Thread
  {
  protected SharedRep srep;
  protected int port;
  protected String dirListing;
  public WebServer(SharedRep srep, int port, String dirListing)
    { this.srep = srep; this.port = port; this.dirListing = dirListing; }
  public void run()
    {
    try
      {
      Properties p = new Properties();
      p.setProperty("port", Integer.toString(port));
      ServletContext sctx = new ServletContext(p, "docroot");
      ServletEngine  seng = new ServletEngine(sctx);
      ServletConfig  scfg = new ServletConfig(new Properties());
      seng.registerServlet("/", new SharedRepServlet(srep, dirListing), scfg);
      seng.serve();
      }
    catch ( IOException e )
      {
      }
    catch ( jakarta.servlet.ServletException e )
      {
      }
    }
  }
//...e
//...sSharedRep:2:
//...sgetProperty:2:
protected static String getProperty(Properties props, String p, String def)
  {
  String v;
  if ( (v = props.getProperty(p)) != null )
    return v;
  return def;
  }
//...e
//...spopulate:2:
// No synchronized blocks, there are no other threads yet

protected void populate(String name)
  throws FileNotFoundException, IOException
  {
  Rep.RepMetaData rmd = rep.getMetaData(name);
  if ( rmd != null )
    {
    if ( rmd.isFile )
      changeMapValue(name, new MapValue(FT_FILE, rmd.loc, rmd.ts));
    else
      changeMapValue(name, new MapValue(FT_TOMBSTONE, rmd.loc, rmd.ts));
    }
  if ( rep.isDirectoryEvenEmpty(name) )
    {
    String[] fns = rep.filesInDirectory(name);
    for ( int i = 0; i < fns.length; i++ )
      populate(name+Rep.SEP+fns[i]);
    String[] tns = rep.tombstonesInDirectory(name);
    for ( int i = 0; i < tns.length; i++ )
      populate(name+Rep.SEP+tns[i]);
    String[] dns = rep.directoriesInDirectory(name);
    for ( int i = 0; i < dns.length; i++ )
      populate(name+Rep.SEP+dns[i]);
    }
  }
//...e

public SharedRep(
  int loc, Properties props,
  SharedRepTrigger trigger
  )
  throws GeneralSecurityException, FileNotFoundException, IOException
  {
  long t0 = System.currentTimeMillis();

  this.loc = loc;
  this.trigger = trigger;
  this.triggerCount = 0;
  this.triggerTime = -1;

  verifyHostname = getProperty(props, "verifyHostname", "true").equalsIgnoreCase("true");
  caseSensitive  = getProperty(props, "caseSensitive" , "true").equalsIgnoreCase("true");
  String dirListing = getProperty(props, "dirListing", "html");

  password  = getProperty(props, "clusterpass" , "password");
  password2 = getProperty(props, "clusterpass2", password  );

  int portAPI  = Integer.parseInt(getProperty(props, "location."+loc+".portAPI" , "6600"));
  int portHTTP = Integer.parseInt(getProperty(props, "location."+loc+".portHTTP", "6700"));

  String basedir = getProperty(props, "location."+loc+".basedir", "content");

  tAdjust = Long.parseLong(getProperty(props, "location."+loc+".tAdjust", "0"));

  rep = new Rep(basedir, loc);
  oSync = new Object();
  map = new HashMap<String,MapValue>();
  locations = new HashMap<Integer,SharedRepLocation>();

  // Setup SSL stuff
  String ksfn = getProperty(props, "location."+loc+".keystore", "SharedRep.jks");
  String kspw = getProperty(props, "location."+loc+".storepass", "password");
  KeyStore ks = KeyStore.getInstance("JKS");
  char[] ksp = kspw.toCharArray();
  ks.load(new FileInputStream(ksfn), ksp);
  KeyManagerFactory kmf = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
  kmf.init(ks, ksp);
  KeyStore ts = KeyStore.getInstance("JKS");
  char[] tsp = kspw.toCharArray();
  ts.load(new FileInputStream(ksfn), tsp);
  TrustManagerFactory tmf = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
  tmf.init(ts);
  SSLContext sc = SSLContext.getInstance("TLS");
  sc.init(kmf.getKeyManagers(), tmf.getTrustManagers(), null);
  sf = sc.getSocketFactory();
  ServerSocketFactory ssf = sc.getServerSocketFactory();
  ss = ssf.createServerSocket(portAPI);

  // Now start processing any fetch requests
  fetching = false;
  fetchq = new java.util.concurrent.LinkedBlockingQueue<FetchRequest>();
  Fetcher f = new Fetcher();
  f.start();

  // Populate our local metadata
  populate("");

  // Set up links with any other locations
  // which we are in a routing group with
  String groupnames = getProperty(props, "groups", "");
  StringTokenizer st = new StringTokenizer(groupnames, ",");
  while ( st.hasMoreTokens() )
    {
    String groupname = st.nextToken();
    String lns = getProperty(props, "group."+groupname, "");
    StringTokenizer st2 = new StringTokenizer(lns, ",");
    Vector<Integer> lnv = new Vector<Integer>();
    while ( st2.hasMoreTokens() )
      lnv.add(Integer.valueOf( Integer.parseInt(st2.nextToken()) ));
    if ( lnv.contains(Integer.valueOf(loc)) )
      for ( int i = 0; i < lnv.size(); i++ )
        {
        int locL = lnv.elementAt(i).intValue();
        if ( locL != loc && locations.get(Integer.valueOf(locL)) == null )
          {
          String hostL = getProperty(props, "location."+locL+".host", "localhost");
          int portL = Integer.parseInt(getProperty(props, "location."+locL+".portAPI", "6600"));
          locations.put(Integer.valueOf(locL), new SharedRepLocation(this, locL, hostL, portL));
          }
        }
    }

  // Now allow API access
  new Server().start();

  // Now allow HTTP access
  new WebServer(this, portHTTP, dirListing).start();

  if ( trigger != null )
    trigger.init();

  int n;
  synchronized ( oSync )
    {
    n = map.size();
    }

  long t1 = System.currentTimeMillis();

  logger.info("SharedRep: version "+VERSION+", "+n+" items, in "+((t1-t0)/1000)+"s, API port "+portAPI+", HTTP port "+portHTTP);

  // Now register ourself as an MBean
  try
    {
    MBeanServer mbs = ManagementFactory.getPlatformMBeanServer(); 
    mbs.registerMBean(this, new ObjectName("nyangau.rep:type=SharedRep"));
    }
  catch ( MalformedObjectNameException e )
    {
    }
  catch ( InstanceAlreadyExistsException e )
    {
    }
  catch ( MBeanRegistrationException e )
    {
    }
  catch ( NotCompliantMBeanException e )
    {
    }
  }
//...e

//...snormalizeName:2:
public static String normalizeName(String name)
  {
  StringTokenizer st = new StringTokenizer(name, "/");
  StringBuffer sb = new StringBuffer();
  while ( st.hasMoreTokens() )
    {
    sb.append(Rep.SEP);
    String s = st.nextToken();
    if ( ! caseSensitive )
      s = s.toLowerCase();
    sb.append(s);
    }
  return sb.toString();
  }
//...e
//...sisFile:2:
public boolean isFile(String name)
  {
  name = normalizeName(name);
  // Note, we could call rep.isFile(name), but its quicker to consult map
  synchronized ( oSync )
    {
    MapValue mvalue = map.get(name);
    return mvalue != null && mvalue.ft == FT_FILE;
    }
  }
//...e
//...sisDirectory:2:
public boolean isDirectory(String name)
  throws FileNotFoundException, IOException
  {
  name = normalizeName(name);
  return rep.isDirectory(name);
  }
//...e
//...sgetLoc:2:
// Not really required for proper operation of SharedRep
// For presenting in web pages etc..

public int getLoc(String name)
  throws FileNotFoundException, IOException
  {
  name = normalizeName(name);
  // Could call rep.getLoc(name), but its quicker to consult mapHere
  synchronized ( oSync )
    {
    MapValue mvalue = map.get(name);
    if ( mvalue == null )
      throw new FileNotFoundException(name);
    return mvalue.loc;
    }
  }
//...e
//...sgetTs:2:
public long getTs(String name)
  throws FileNotFoundException, IOException
  {
  name = normalizeName(name);
  // Could call rep.getTs(name), but its quicker to consult map
  synchronized ( oSync )
    {
    MapValue mvalue = map.get(name);
    if ( mvalue == null )
      throw new FileNotFoundException(name);
    return mvalue.ts;
    }
  }
//...e
//...sgetLength:2:
protected long getLength(String name)
  throws FileNotFoundException, IOException
  {
  name = normalizeName(name);
  return rep.getLength(name);
  }
//...e
//...sfilesInDirectory:2:
public String[] filesInDirectory(String name)
  throws FileNotFoundException, IOException
  {
  name = normalizeName(name);
  return rep.filesInDirectory(name);
  }
//...e
//...sdirectoriesInDirectory:2:
public String[] directoriesInDirectory(String name)
  throws FileNotFoundException, IOException
  {
  name = normalizeName(name);
  return rep.directoriesInDirectory(name);
  }
//...e
//...sreadFile:2:
// This one is intended for external callers to use
// Gets the current version of the file

public byte[] readFile(String name)
  throws FileNotFoundException, IOException
  {
  name = normalizeName(name);
  return rep.readFile(name);
  }

// This one exists primarily to ensure that fetch requests will be
// sure to get the exact version they want, ie: (locWanted, tsWanted) tuple

public byte[] readFile(String name, int locWanted, long tsWanted)
  throws FileNotFoundException, IOException
  {
  name = normalizeName(name);
  synchronized ( oSync )
    {
    // Could call rep.getMetaData(name)
    // but its quicker to consult map
    MapValue mvalue = map.get(name);
    if ( mvalue == null )
      throw new FileNotFoundException(name);
    if ( mvalue.ft != FT_FILE )
      throw new FileNotFoundException(name+" is not a file");
    if ( locWanted != -1 && mvalue.loc != locWanted )
      throw new FileNotFoundException(name+" has wrong loc");
    if ( tsWanted != -1 && mvalue.ts != tsWanted )
      throw new FileNotFoundException(name+" has wrong ts");
    return rep.readFile(name);
    }
  }
//...e
//...swriteFile:2:
public void writeFile(String name, boolean overwrite, byte[] b)
  throws FileNotFoundException, IOException
  {
  name = normalizeName(name);
  synchronized ( oSync )
    {
    MapValue mvalueOld = map.get(name);
    long ts;
    MapValue mvalue;
    for ( ;; )
      {
      ts = System.currentTimeMillis()+getTimeAdjustment();
      mvalue = new MapValue(FT_FILE, this.loc, ts);
      if ( mvalueOld == null || mvalueOld.compareTo(mvalue) != 0 )
        break;
      // We're trying to overwrite our own work very quickly
      // Let time move on so we are sure to overwrite the prior version
      try
        {
        Thread.sleep(5);
        }
      catch ( InterruptedException e )
        {
        }
      }
    if ( mvalueOld == null ||
         mvalueOld.compareTo(mvalue) < 0 )
      {
      rep.writeFile(name, this.loc, ts, overwrite, b);
      changeMapValue(name, mvalue);
      }
    if ( trigger != null )
      trigger.written(name, b);
    ++triggerCount;
    triggerTime = System.currentTimeMillis();
    }
  }
//...e
//...sdeleteFile:2:
public void deleteFile(String name)
  throws FileNotFoundException, IOException
  {
  name = normalizeName(name);
  synchronized ( oSync )
    {
    MapValue mvalueOld = map.get(name);
    long ts;
    MapValue mvalue;
    for ( ;; )
      {
      ts = System.currentTimeMillis()+getTimeAdjustment();
      mvalue = new MapValue(FT_TOMBSTONE, this.loc, ts);
      if ( mvalueOld == null || mvalueOld.compareTo(mvalue) != 0 )
        break;
      // We're trying to overwrite our own work very quickly
      // Let time move on so we are sure to overwrite the prior version
      try
        {
        Thread.sleep(5);
        }
      catch ( InterruptedException e )
        {
        }
      }
    if ( mvalueOld == null ||
         mvalueOld.compareTo(mvalue) < 0 )
      {
      rep.deleteFile(name, this.loc, ts);
      changeMapValue(name, mvalue);
      }
    if ( trigger != null )
      trigger.deleted(name);
    ++triggerCount;
    triggerTime = System.currentTimeMillis();
    }
  }
//...e
//...sfindXPath:2:
//...sclass NsCtx:2:
public class NsCtx implements NamespaceContext
  {
  protected Map<String,String> map;
//...sclass RoIterator:4:
protected class RoIterator implements Iterator<String>
  {
  protected Iterator<String> iter;
  public RoIterator(Iterator<String> i)
    { iter = i; }
  public boolean hasNext()
    { return iter.hasNext(); };
  public String next()
    { return iter.next(); }
  public void remove()
    throws UnsupportedOperationException
    { throw new UnsupportedOperationException(); }
  }
//...e
//...sNsCtx:4:
// Likely to be followed by addPrefix() calls
public NsCtx()
  {
  map = new HashMap<String,String>();
  map.put(XMLConstants.DEFAULT_NS_PREFIX, XMLConstants.NULL_NS_URI           );
  map.put(XMLConstants.XML_NS_PREFIX    , XMLConstants.XML_NS_URI            );
  map.put(XMLConstants.XMLNS_ATTRIBUTE  , XMLConstants.XMLNS_ATTRIBUTE_NS_URI);
  }

// Typical use
public NsCtx(Map<String,String> m)
  {
  this();
  map.putAll(m);
  }
//...e
//...saddPrefix:4:
public void addPrefix(String prefix, String namespaceURI)
  {
  map.put(prefix, namespaceURI);
  }
//...e
//...sgetNamespaceURI:4:
public String getNamespaceURI(String prefix)
  {
  if ( prefix == null )
    throw new IllegalArgumentException();
  return map.get(prefix);
  }
//...e
//...sgetPrefix:4:
public String getPrefix(String namespaceURI)
  {
  if ( namespaceURI == null )
    throw new IllegalArgumentException();
  for ( String k : map.keySet() )
    if ( map.get(k).equals(namespaceURI) )
      return k;
  return null;
  }
//...e
//...sgetPrefixes:4:
public Iterator<String> getPrefixes(String namespaceURI)
  {
  if ( namespaceURI == null )
    throw new IllegalArgumentException();
  Set<String> prefixes = new HashSet<String>();
  for ( String k : map.keySet() )
    if ( map.get(k).equals(namespaceURI) )
      prefixes.add(k);
  return new RoIterator(prefixes.iterator());
  }
//...e
  }
//...e

protected static final String NS_REP = "http://www.nyangau.org/rep";
protected static final String PREFIX_REP = "r";

//...sbytesToString:2:
//...sXML scanning helpers:2:
// Code used for scanning <?xml text, looking for encoding="XYZ"

//...swsChar:2:
protected static boolean wsChar(char c)
  {
  return c == ' ' || c == '\t' || c == '\n' || c == '\r';
  }
//...e
//...snameStartChar:2:
protected static boolean nameStartChar(char c)
  {
  return ( c >= 'A' && c <= 'Z' ) ||
         ( c >= 'a' && c <= 'z' ) ||
         c == ':' || c == '_' ||
         ( c >= '\u00c0' && c <= '\u00d6' ) ||
         ( c >= '\u00d8' && c <= '\u00f6' ) ||
         ( c >= '\u00f8' && c <= '\u02ff' ) ||
         ( c >= '\u0370' && c <= '\u037d' ) ||
         ( c >= '\u037f' && c <= '\u1fff' ) ||
         ( c >= '\u200c' && c <= '\u200d' ) ||
         ( c >= '\u2070' && c <= '\u218f' ) ||
         ( c >= '\u2c00' && c <= '\u2fef' ) ||
         ( c >= '\u3001' && c <= '\ud7ff' ) ||
         ( c >= '\uf900' && c <= '\ufdcf' ) ||
         ( c >= '\ufdf0' && c <= '\ufffd' );
  }
//...e
//...snameChar:2:
protected static boolean nameChar(char c)
  {
  return nameStartChar(c) ||
         ( c >= '0' && c <= '9' ) ||
         c == '-' || c == '.' ||
         c == (char)0xb7 ||
         ( c >= '\u0300' && c <= '\u036f' ) ||
         ( c >= '\u203f' && c <= '\u2040' );
  }
//...e

// Two almost identical implementations of nextToken, so as to avoid the
// need to convert between String <-> byte[] before using it.

//...snextToken byte\91\\93\:2:
protected static int nextToken(byte[] b, int i, StringBuffer sb)
  {
  while ( i < b.length && wsChar((char)b[i]) )
    i++;
  if ( i == b.length )
    return -1; // no more tokens
  if ( (char)b[i] == '?' && i+2 <= b.length && (char)b[i+1] == '>' )
    return -1; // no more tokens, as we hit ?>
  if ( nameStartChar((char)b[i]) )
    {
    sb.append((char)b[i++]);
    while ( i < b.length && nameChar((char)b[i]) )
      sb.append((char)b[i++]);
    return i;
    }
  if ( (char)b[i] == '"' )
    {
    ++i;
    sb.append("\""); // so caller knows its a quoted string
    while ( i <= b.length && (char)b[i] != '"' )
      sb.append((char)b[i++]);
    if ( i == b.length )
      return -1; // unterminated string
    return i+1;
    }
  else if ( (char)b[i] == '=' )
    {
    sb.append('=');
    return i+1;
    }
  else
    return -1; // something unexpected
  }
//...e
//...e

public static String bytesToString(byte[] b, String enc)
  throws IOException
  {
  int skip = 0;
  if ( enc == null )
    enc = "UTF-8";
  if ( enc.equalsIgnoreCase("XML") )
    // Try to autodetect encoding used, based on BOM or XML header
    {
    if ( b.length >= 3 &&
         b[0] == (byte) 0xef && b[1] == (byte) 0xbb && b[2] == (byte) 0xbf )
      {
      enc = "UTF-8";
      skip = 3;
      }
    else if ( b.length >= 2 &&
         b[0] == (byte) 0xfe && b[1] == (byte) 0xff )
      {
      enc = "UTF-16BE";
      skip = 2;
      }
    else if ( b.length >= 2 &&
         b[0] == (byte) 0xff && b[1] == (byte) 0xfe )
      {
      enc = "UTF-16LE";
      skip = 3;
      }
    else if ( b.length >= 6 &&
         b[0] == '<' && b[1] == '?' && b[2] == 'x' && b[3] == 'm' && b[4] == 'l' &&
         wsChar((char)b[5]) )
      // No BOM, but it looks like XML, so try scanning the <?xml text
      {
      String token2 = "", token1 = "", token0 = "";
      int i = 6;
      for ( ;; )
        {
        StringBuffer sb = new StringBuffer();
        i = nextToken(b, i, sb);
        if ( i == -1 )
          {
          enc = "UTF-8";
          break;
          }
        token2 = token1; token1 = token0; token0 = sb.toString();
        if ( token2.equals("encoding") && token1.equals("=") && token0.startsWith("\"") )
          {
          enc = token0.substring(1);
          if ( enc.equals("UTF-16") )
            throw new IOException("can't convert byte[] to String using encoding XML, there was no BOM, but we found <?xml encoding=\"UTF-16\", which is illegal");
          break;
          }
        }
      }
    else
      // No BOM, not XML, might as well try UTF-8
      enc = "UTF-8";
    }
  try
    {
    return new String(b, skip, b.length-skip, enc);
    }
  catch ( UnsupportedEncodingException e )
    {
    throw new IOException("can't convert byte[] to String using encoding "+enc+": "+e.getMessage());
    }
  }
//...e
//...sstrToDoc:2:
protected static Document strToDoc(String s)
  throws SAXException
  {
  DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
  // not validating
  dbf.setNamespaceAware(true); // must do this, else can't validate ok
  dbf.setCoalescing(true);
  dbf.setExpandEntityReferences(true);
  dbf.setIgnoringComments(false);
  try
    {
    DocumentBuilder db = dbf.newDocumentBuilder();
    db.setErrorHandler(
      new ErrorHandler()
        {
        public void warning(SAXParseException e)
          throws SAXException
          {
          }
        public void error(SAXParseException e)
          throws SAXException
          {
          }
        public void fatalError(SAXParseException e)
          throws SAXException
          {
          }
        }
      );
    Document d = db.parse(new InputSource(new StringReader(s)));
      // Note: we avoid using new InputStream(new StringInputStream()),
      // as we want to avoid working at the byte level,
      // after all, we already have a String, which is a sequence of chars.
      // Note that StringInputStream is deprecated for this reason.
    return d;
    }
  catch ( ParserConfigurationException e )
    {
    logger.severe("SharedRep: strToDoc: "+e.getMessage());
    return null;
    }
  catch ( IOException e )
    {
    logger.severe("SharedRep: strToDoc: "+e.getMessage());
    return null;
    }
  }
//...e
//...sdocToStr:2:
protected static String docToStr(Document d)
  {
  try
    {
    TransformerFactory tf = TransformerFactory.newInstance();
    Transformer t = tf.newTransformer(); // the "identity" transformation
    t.setOutputProperty(OutputKeys.INDENT, "yes");
    CharArrayWriter caw = new CharArrayWriter();
    t.transform(
      new DOMSource((Node) d),
      new StreamResult(caw)
      );
    caw.flush();
    return caw.toString();
    }
  catch ( TransformerException e )
    {
    logger.severe("SharedRep: docToStr: "+e.getMessage());
    }
  return null;
  }
//...e
//...sfindXPathFile:2:
protected void findXPathFile(
  String name,
  Pattern pattern,
  boolean content,
  XPathExpression xpe,
  Document dRoot,
  Element eRoot
  )
  throws FileNotFoundException, IOException, XPathExpressionException
  {
  // Check against pattern
  if ( pattern != null )
    {
    int i = name.lastIndexOf(Rep.SEP);
    Matcher m = pattern.matcher(name.substring(i+1));
    if ( ! m.matches() )
      return;
    }
  // Read
  byte b[];
  try
    {  
    b = readFile(name);
    }
  catch ( FileNotFoundException e )
    {
    return;
    }
  catch ( IOException e )
    {
    Element eFile = dRoot.createElementNS(NS_REP, "fileError");
    eFile.setPrefix(PREFIX_REP);
    eFile.setAttribute("name", name);
    eFile.setAttribute("info", "reading");
    eRoot.appendChild(eFile);
    return;
    }
  // Decode
  String s;
  try
    {
    s = bytesToString(b, "XML");
    }
  catch ( IOException e )
    {
    Element eFile = dRoot.createElementNS(NS_REP, "fileError");
    eFile.setPrefix(PREFIX_REP);
    eFile.setAttribute("name", name);
    eFile.setAttribute("info", "encoding");
    eRoot.appendChild(eFile);
    return;
    }
  // Parse
  Document d;
  try
    {
    d = strToDoc(s);
    }
  catch ( SAXException e )
    {
    Element eFile = dRoot.createElementNS(NS_REP, "fileError");
    eFile.setPrefix(PREFIX_REP);
    eFile.setAttribute("name", name);
    eFile.setAttribute("info", "parsing");
    eRoot.appendChild(eFile);
    return;
    }
  // Scan
  NodeList nl = (NodeList) xpe.evaluate(d, XPathConstants.NODESET);
  if ( nl.getLength() != 0 )
    {
    Element eFile = dRoot.createElementNS(NS_REP, "file");
    eFile.setPrefix(PREFIX_REP);
    eFile.setAttribute("name", name);
    eRoot.appendChild(eFile);
    if ( content )
      {
      for ( int i = 0; i < nl.getLength(); i++ )
        {
        Node n = nl.item(i);
        eFile.appendChild(dRoot.importNode(n, true));
        }
      }
    }
  }
//...e
//...sfindXPathDirectory:2:
protected void findXPathDirectory(
  String name,
  Pattern pattern,
  boolean recurse,
  boolean content,
  XPathExpression xpe,
  Document dRoot,
  Element eRoot
  )
  throws FileNotFoundException, IOException, XPathExpressionException
  {
  String[] fns = filesInDirectory(name);
  for ( int i = 0; i < fns.length; i++ )
    findXPathFile(name+"/"+fns[i], pattern, content, xpe, dRoot, eRoot);
  if ( recurse )
    {
    String[] dns = directoriesInDirectory(name);
    for ( int i = 0; i < dns.length; i++ )
      findXPathDirectory(name+"/"+dns[i], pattern, recurse, content, xpe, dRoot, eRoot);
    }
  }
//...e

public String findXPath(
  String name,      // can be file or directory
  String filter,    // can be regexp to match against file basename, or null
  boolean recurse,  // do subdirectories too
  boolean content,  // just list matching files, or include content too
  String xpath,     // XPath expression to use
  String nsprefixes // Namespace prefixes, or null
  )
  throws FileNotFoundException, IOException, PatternSyntaxException, XPathExpressionException
  {
  name = normalizeName(name);
  Pattern pattern = ( filter != null ) ? Pattern.compile(filter) : null;

  XPath xp = XPathFactory.newInstance().newXPath();
  if ( nsprefixes != null )
    {
    NsCtx nsctx = new NsCtx();
    StringTokenizer st = new StringTokenizer(nsprefixes);
    while ( st.hasMoreTokens() )
      {
      String s2 = st.nextToken();
      int i = s2.indexOf("=");
      if ( i == -1 )
        throw new XPathExpressionException("bad namespace context");
      nsctx.addPrefix(s2.substring(0, i), s2.substring(i+1));
      }
    xp.setNamespaceContext(nsctx);
    }
  XPathExpression xpe = xp.compile(xpath);

  DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
  dbf.setNamespaceAware(true);
  dbf.setCoalescing(true);
  dbf.setExpandEntityReferences(true);
  dbf.setIgnoringComments(false);
  try
    {
    DocumentBuilder db = dbf.newDocumentBuilder();
    Document d = db.newDocument();
    Element eRoot = d.createElementNS(NS_REP, "files");
    eRoot.setPrefix(PREFIX_REP);
    d.appendChild(eRoot);
    if ( isFile(name) )
      findXPathFile(name, pattern, content, xpe, d, eRoot);
    else if ( isDirectory(name) )
      findXPathDirectory(name, pattern, recurse, content, xpe, d, eRoot);
    return docToStr(d);
    }
  catch ( ParserConfigurationException e )
    {
    logger.severe("SharedRep: findXPath: "+e.getMessage());
    }
  return "";
  }
//...e

//...sgetBacklog:2:
// Not really required for proper operation of SharedRep
// For presenting in web pages etc..
// Also part of SharedRepMBean interface
// Add one to account for a fetch in progress

public int getBacklog()
  {
  return fetchq.size() + ( fetching ? 1 : 0 );
  }
//...e
//...sgetTs:2:
// Not really required for proper operation of SharedRep
// Part of SharedRepMBean interface

public long getTs()
  { return System.currentTimeMillis()+getTimeAdjustment(); }
//...e
//...sgetMapSize:2:
// Not really required for proper operation of SharedRep
// Part of SharedRepMBean interface

public int getMapSize()
  {
  synchronized ( oSync )
    {
    return map.size();
    }
  }
//...e
//...sgetChangeCount:2:
// Part of SharedRepMBean interface
public int getChangeCount()
  { return triggerCount; }
//...e
//...sgetLastChangeTime:2:
// Part of SharedRepMBean interface
public String getLastChangeTime()
  {
  return ( triggerTime != -1 )
    ? new Date(triggerTime).toString()
    : "";
  }
//...e
//...stoString:2:
public String toString()
  {
  String m;
  synchronized ( oSync )
    {
    m = map.toString();
    }
  return "{loc="+loc+",map="+m+",locations="+locations+"}";
  }
//...e
//...sdebugCommand:2:
public boolean debugCommand(String[] args)
  throws FileNotFoundException, IOException, XPathExpressionException
  {
  if ( args[0].equals("dump") )
    rep.dump();
  else if ( args[0].equals("toString") )
    System.out.println(toString());
  else if ( args[0].equals("getBacklog") )
    System.out.println(getBacklog());
  else if ( args[0].equals("readFile") )
    {
    byte[] b = readFile(args[1]);
    System.out.write(b);
    }
  else if ( args[0].equals("writeFile") )
    {
    byte[] b = (args[2]+"\n").getBytes("UTF-8");
    writeFile(args[1], true, b);
    }
  else if ( args[0].equals("deleteFile") )
    deleteFile(args[1]);
  else if ( args[0].equals("findXPath") )
    System.out.println(
      findXPath(
        args[1],
        args[2].equals("-") ? null : args[2],
        args[3].equalsIgnoreCase("true"),
        args[4].equalsIgnoreCase("true"),
        args[5],
        args[6].equals("-") ? null : args[6].replace(',',' ')
        )
      );
  else if ( args[0].equals("help") )
    {
    System.out.println("dump");
    System.out.println("toString");
    System.out.println("getBacklog");
    System.out.println("readFile name");
    System.out.println("writeFile name data");
    System.out.println("deleteFile name");
    System.out.println("findXPath name filter recurse content xpath nsprefixes");
    }
  else
    return false;
  return true;
  }
//...e
//...sdebugDriver:2:
// Interactive test driver

public void debugDriver()
  throws FileNotFoundException, IOException
  {
  InputStreamReader isr = new InputStreamReader(System.in);
  BufferedReader br = new BufferedReader(isr);
  String line;
  while ( (line = br.readLine()) != null )
    {
    StringTokenizer st = new StringTokenizer(line);
    Vector<String> v = new Vector<String>();
    while ( st.hasMoreTokens() )
      v.add(st.nextToken());
    if ( v.size() > 0 )
      {
      String[] a = v.toArray(new String[0]);
      try
        {
        if ( ! debugCommand(a) )
          System.out.println("Not understood, use help");
        }
      catch ( Exception e )
        {
        e.printStackTrace();
        }
      }
    }
  }
//...e
//...smain:2:
//...susage:2:
protected static void usage()
  {
  System.out.println("usage: SharedRep [flags]");
  System.out.println("flags: -l loc           unique location number of this instance");
  System.out.println("       -c props         cluster properties file (default: cluster.props)");  
  System.out.println("       -t triggerclass  classname to trigger (default: none)");
  System.exit(1);
  }
//...e

public static void main(String[] args)
  {
  try
    {
    int loc = -1;
    String propsfn = "cluster.props";
    String pwServer = "password";
    String triggerclass = null;
    boolean debug = false;
    SharedRepTrigger trigger;

//...sparse arguments:6:
int i;
for ( i = 0; i < args.length; i++ )
  if ( args[i].charAt(0) != '-' )
    break;
  else if ( args[i].charAt(1) == '-' )
    { ++i; break; }
  else
    switch ( args[i].charAt(1) )
      {
      case 'l':
        if ( ++i == args.length )
          usage();
        loc = Integer.parseInt(args[i]);
        break;
      case 'c':
        if ( ++i == args.length )
          usage();
        propsfn = args[i];
        break;
      case 't':
        if ( ++i == args.length )
          usage();
        triggerclass = args[i];
        break;
      case 'd':
        debug = true;
        break;
      default:
        usage();
        break;
      }

if ( i != args.length )
  usage();

if ( loc == -1 )
  usage();

if ( triggerclass != null )
  {
  Class<?> c = Class.forName(triggerclass);
  Constructor ctor = c.getConstructor(new Class[]{});
  trigger = (SharedRepTrigger) ctor.newInstance(new Object[]{});
  }
else
  trigger = null;
//...e

    Properties props = new Properties();
    FileInputStream fis = new FileInputStream(propsfn);
    props.load(fis);
    fis.close();

    SharedRep srep = new SharedRep(loc, props, trigger);

    if ( debug )
      {
      srep.debugDriver();
      System.exit(0);
      }
    }
  catch ( Exception e )
    {
    e.printStackTrace();
    }
  }
//...e
  }
