//
// SharedRepServlet.java - Servlet to serve SharedRepository content
//

package nyangau.rep;

//...simports:0:
import java.net.*;
import java.io.*;
import java.util.*;
import java.net.URL;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import jakarta.servlet.*;
import jakarta.servlet.http.*;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.stream.StreamResult;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.Text;
//...e

public class SharedRepServlet extends HttpServlet
  {
//...sgetInitParameter:2:
protected String getInitParameter(String p)
  {
  ServletConfig sc = getServletConfig();
  return sc.getInitParameter(p);
  }

protected String getInitParameter(String p, String def)
  {
  String v;
  if ( (v = getInitParameter(p)) != null )
    return v;
  else
    return def;
  }
//...e
  protected Dictionary<String,String> icons = new Hashtable<String,String>();
  protected SharedRep srep;
  protected boolean xmlDef;
//...ssetupIcons:2:
// A selection of mappings from file extension to icons
// Can be overridden or extended in properties

protected void setupIcons()
  {
  icons.put("doc",    "word.gif");
  icons.put("xls",    "excel.gif");
  icons.put("ppt",    "powerpoint.gif");
  icons.put("mpp",    "project.gif");
  icons.put("pdf",    "acrobat.gif");
  icons.put("ps" ,    "postscript.gif");
  icons.put("zip",    "archive.gif");
  icons.put("tar",    "archive.gif");
  icons.put("au",     "audio.gif");
  icons.put("snd",    "audio.gif");
  icons.put("aif",    "audio.gif");
  icons.put("aiff",   "audio.gif");
  icons.put("aifc",   "audio.gif");
  icons.put("wav",    "audio.gif");
  icons.put("gif",    "image.gif");
  icons.put("png",    "image.gif");
  icons.put("jpg",    "image.gif");
  icons.put("jpe",    "image.gif");
  icons.put("jpeg",   "image.gif");
  icons.put("tif",    "image.gif");
  icons.put("tiff",   "image.gif");
  icons.put("pnm",    "image.gif");
  icons.put("pbm",    "image.gif");
  icons.put("pgm",    "image.gif");
  icons.put("ppm",    "image.gif");
  icons.put("xbm",    "image.gif");
  icons.put("xpm",    "image.gif");
  icons.put("xwd",    "image.gif");
  icons.put("htm",    "text.gif");
  icons.put("html",   "text.gif");
  icons.put("txt",    "text.gif");
  icons.put("text",   "text.gif");
  icons.put("mpg",    "video.gif");
  icons.put("mpe",    "video.gif");
  icons.put("mpeg",   "video.gif");
  icons.put("qt",     "video.gif");
  icons.put("mov",    "video.gif");
  icons.put("avi",    "video.gif");
  icons.put("movie",  "video.gif");
  }
//...e
//...sSharedRepServlet:2:
public SharedRepServlet(SharedRep srep, String dirListing)
  {
  this.srep = srep;
  xmlDef = dirListing.equalsIgnoreCase("xml");
  }
//...e
//...ssendHeader:2:
protected void sendHeader(HttpServletResponse resp, String title)
  throws IOException
  {
  resp.setContentType("text/html");
  ServletOutputStream os = resp.getOutputStream();
  os.println("<HTML>");
  os.println("<HEAD>");
  os.println("<TITLE>"+title+"</TITLE>");
  os.println("</HEAD>");
  os.println("<BODY BGCOLOR=\"#FFFFFF\">");
  os.println("<H1>"+title+"</H1>");
  os.println();
  }
//...e
//...ssendTrailer:2:
protected void sendTrailer(HttpServletResponse resp)
  throws IOException
  {
  ServletOutputStream os = resp.getOutputStream();
  os.println();
  os.println("</BODY>");
  os.println("</HTML>");
  }
//...e
//...ssendFile:2:
//...sgetMimeType:2:
// If the file has a file extension, try to determine MIME type by
//   looking to see if we've got a mime.ext property
//   looking to see if the the container can map the filename to a MIME type

public String getMimeType(String fn)
  {
  String bfn = fn.substring(fn.lastIndexOf('/')+1);
  int dot;
  if ( (dot = bfn.lastIndexOf('.')) != -1 )
    {
    String ext = bfn.substring(dot+1);
    String type;
    if ( (type = getInitParameter("mime."+ext)) != null )
      return type;
    ServletContext sc = getServletConfig().getServletContext();
    if ( (type = sc.getMimeType(fn)) != null )
      return type;
    }
  return null;
  }
//...e

protected void sendFile(
  HttpServletRequest req,
  HttpServletResponse resp,
  String name
  )
  throws IOException
  {
  String locS = req.getParameter("loc");
  int locWanted = ( locS != null ) ? Integer.parseInt(locS) : -1;
  String tsS = req.getParameter("ts");
  long tsWanted = ( tsS != null ) ? Long.parseLong(tsS) : -1;
  String mime = req.getParameter("mime");
  if ( mime != null )
    ; // user has chosen
  else if ( (mime = getMimeType(name)) != null )
    ; // we worked it out
  else
    mime = "text/plain";
  resp.setContentType(mime);
  try
    {
    byte[] b = srep.readFile(name, locWanted, tsWanted);
    ServletOutputStream os = resp.getOutputStream();
    os.write(b);
    }
  catch ( FileNotFoundException e )
    {
    resp.sendError(HttpServletResponse.SC_NOT_FOUND);
    }
  catch ( IOException e )
    {
    resp.sendError(HttpServletResponse.SC_NOT_FOUND);
    }
  }
//...e
//...ssendDirectory:2:
//...sgetIcon:2:
// If the file has a file extension, try to determine icon
//   looking to see if the the container can map the filename to a MIME type

public String getIcon(String fn)
  {
  String bfn = fn.substring(fn.lastIndexOf('/')+1);
  int dot;
  if ( (dot = bfn.lastIndexOf('.')) != -1 )
    {
    String ext = bfn.substring(dot+1);
    String icon;
    if ( (icon = getInitParameter("icon."+ext)) != null )
      return icon;
    if ( (icon = icons.get(ext)) != null )
      return "/.images/"+icon;
    }
  return null;
  }
//...e
//...sencodeHTML:2:
// For encoding data within HTML output.
// This doesn't try to be clever.
// It could try harder to map to HTML entities with names, such as &cent;
// It just does the common ones, and falls back to numeric for the rest.

protected String encodeHTML(String s)
  {
  StringBuilder sb = new StringBuilder( s.length() );
  for ( int i = 0; i < s.length(); i++ )
    {
    char c = s.charAt(i);
    switch ( c )
      {
      case '<': sb.append("&lt;"); break;
      case '>': sb.append("&gt;"); break;
      case '&': sb.append("&amp;"); break;
      case '"': sb.append("&quot;"); break;
      default:
        if ( c >= ' ' && c <= '~' )
          // regular printable ASCII character
          sb.append(c);
        else
          // use numeric entity notation for the rest
          {
          sb.append("&#");
          sb.append(c);
          sb.append(";");
          }
        break;
      }
    }
  return sb.toString();
  }
//...e

protected void sendDirectory(
  HttpServletRequest req,
  HttpServletResponse resp,
  String name,
  boolean xml
  )
  throws IOException
  {
  long tsAdjust = srep.getTimeAdjustment();
  String meta = req.getParameter("meta"); // show metadata
  if ( xml )
//...sin XML form:6:
{
resp.setContentType("text/xml");
ServletOutputStream os = resp.getOutputStream();

DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
dbf.setNamespaceAware(true);
dbf.setCoalescing(true);
dbf.setExpandEntityReferences(true);
dbf.setIgnoringComments(false);
try
  {
  DocumentBuilder db = dbf.newDocumentBuilder();
  Document d = db.newDocument();
  Element eRoot = d.createElement("directory");
  eRoot.setAttribute("loc", Integer.toString(srep.getLoc()));
  eRoot.setAttribute("ts", Long.toString(srep.getTs()));
  d.appendChild(eRoot);

  String[] dns = srep.directoriesInDirectory(name);
  Arrays.sort(dns);
  for ( int i = 0; i < dns.length; i++ )
    {
    String dn = dns[i];
    String n = name+"/"+dn;
    Element e = d.createElement("subdirectory");
    e.appendChild(d.createTextNode(dn));
    eRoot.appendChild(e);
    }

  String[] fns = srep.filesInDirectory(name);
  Arrays.sort(fns);
  for ( int i = 0; i < fns.length; i++ )
    {
    String fn = fns[i];
    String n = name+"/"+fn;
    Element e = d.createElement("file");
    long ts = srep.getTs(n);
    if ( meta != null )
      {
      e.setAttribute("loc", Integer.toString(srep.getLoc(n)));
      e.setAttribute("ts", Long.toString(ts));
      }
    e.setAttribute("length", Long.toString(srep.getLength(n)));
    e.setAttribute("localts", Long.toString(ts-tsAdjust));
    e.appendChild(d.createTextNode(fn));
    eRoot.appendChild(e);
    }

  TransformerFactory tf = TransformerFactory.newInstance();
  Transformer t = tf.newTransformer(); // the "identity" transformation
  t.setOutputProperty(OutputKeys.INDENT, "yes");
  t.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
  t.setOutputProperty("{http://saxon.sf.net/}indent-spaces", "4");
  t.transform(
    new DOMSource((Node) d),
    new StreamResult(os)
    );
  }
catch ( ParserConfigurationException e )
  {
  }
catch ( TransformerException e )
  {
  }

}
//...e
  else
//...sin pretty HTML form:6:
{
String dirSuffix = xmlDef ? "?html" : "";
String title = "SharedRep";
sendHeader(resp, title+" - "+encodeHTML(name)+"/");
ServletOutputStream os = resp.getOutputStream();
os.println("<P>");
os.println("<TABLE>");
os.print("<TR><TD BGCOLOR=\"#CCCCFF\">Icon<TD BGCOLOR=\"#CCCCFF\">Name<TD ALIGN=RIGHT BGCOLOR=\"#CCCCFF\">Size<TD BGCOLOR=\"#CCCCFF\">Local Ts");
if ( meta != null )
  os.print("<TD ALIGN=RIGHT BGCOLOR=\"#CCCCFF\">Loc<TD ALIGN=RIGHT BGCOLOR=\"#CCCCFF\">Ts");
os.println();
String[] bg = { " BGCOLOR=\"#EEEEEE\"", " BGCOLOR=\"#DDDDDD\"" };
int j = 0;
if ( ! name.equals("") )
  {
  os.print("<TR><TD"+bg[0]+"><IMG SRC=\"/.images/up.gif\" ALT=\"\"><TD"+bg[0]+"><A HREF=\".."+dirSuffix+"\">..</A><TD ALIGN=RIGHT"+bg[0]+">-<TD"+bg[0]+">-");
  if ( meta != null )
    os.print("<TD ALIGN=RIGHT"+bg[0]+">-<TD ALIGN=RIGHT"+bg[0]+">-");
  os.println();
  j++;
  }

String[] dns = srep.directoriesInDirectory(name);
Arrays.sort(dns);
for ( int i = 0; i < dns.length; i++ )
  {
  String dn = dns[i];
  String dnH = encodeHTML(dn);
//  String n = name+"/"+dn;
  os.print("<TR><TD"+bg[j&1]+"><IMG SRC=\"/.images/directory.gif\" ALT=\"\">"+
               "<TD"+bg[j&1]+"><A HREF=\""+resp.encodeURL(dnH+"/")+dirSuffix+"\">"+dnH+"</A>"+
               "<TD"+bg[j&1]+" ALIGN=RIGHT>-<TD "+bg[j&1]+">-");
  if ( meta != null )
    os.print("<TD ALIGN=RIGHT"+bg[j&1]+">-<TD ALIGN=RIGHT"+bg[j&1]+">-");
  os.println();
  j++;
  }  

String[] fns = srep.filesInDirectory(name);
Arrays.sort(fns);
for ( int i = 0; i < fns.length; i++ )
  {
  String fn = fns[i];
  String fnH = encodeHTML(fn);
  String n = name+"/"+fn;
  String icon;
  if ( (icon = getIcon(fn)) == null )
    icon = "/.images/unknown.gif";
  long ts = srep.getTs(n);
  os.print("<TR><TD"+bg[j&1]+"><IMG SRC=\""+icon+"\" ALT=\"\">"+
               "<TD"+bg[j&1]+"><A HREF=\""+resp.encodeURL(fnH)+"\">"+fnH+"</A>"+
               "<TD"+bg[j&1]+" ALIGN=RIGHT>"+srep.getLength(n)+
               "<TD"+bg[j&1]+">"+new Date(ts-tsAdjust));
    if ( meta != null )
      os.print("<TD ALIGN=RIGHT"+bg[j&1]+">"+srep.getLoc(n)+"<TD ALIGN=RIGHT"+bg[j&1]+">"+new Date(ts));
  os.println();
  j++;
  }

os.println("</TABLE>");

os.println("<P>Backlog: "+srep.getBacklog()+
           ", Loc: "+srep.getLoc()+
           ", Ts: "+new Date( System.currentTimeMillis()+srep.getTimeAdjustment() )
          );

sendTrailer(resp);
}
//...e
  }
//...e
//...sdoGet:2:
protected void doGet(HttpServletRequest req, HttpServletResponse resp)
  throws ServletException, IOException
  {
  String uri = req.getRequestURI();
  uri = URLDecoder.decode(uri, "UTF-8");
  String name = SharedRep.normalizeName(uri);
  if ( srep.isFile(name) && ! uri.endsWith("/") )
    sendFile(req, resp, name);
  else if ( srep.isDirectory(name) || name.equals("") )
    {
    boolean xml;
    String dirSuffix;
    if ( req.getParameter("xml") != null )
      { xml = true; dirSuffix = "?xml"; }
    else if ( req.getParameter("html") != null )
      { xml = false; dirSuffix = "?html"; }
    else
      { xml = xmlDef; dirSuffix = ""; }
    if ( uri.endsWith("/") )
      sendDirectory(req, resp, name, xml);
    else
      resp.sendRedirect(resp.encodeRedirectURL(uri+"/")+dirSuffix);
    }
  else      
    resp.sendError(HttpServletResponse.SC_NOT_FOUND);
  }
//...e
//...sinit:2:
public void init()
  throws ServletException
  {
  super.init();
  setupIcons();
  }
//...e
  }
