#!/bin/sh
#
# rep.sh - Shared Repository control script
#
# Full path to cluster.props included in the command, although its in the
# current directory, so that when we kill it, we kill the right process.
#

java_home=/usr/lib/jvm/java-11-openjdk
rep_home=/work/ak/java/rep
loc=0
cluster=$rep_home/cluster.props

#

cd $rep_home

cmd="nyangau.rep.SharedRep -l $loc -c $cluster"
jcmd="$java_home/bin/java -Xmx256m -cp nyangau-rep.jar:nyangau-se.jar $cmd"

case `uname -s` in
	Linux|Darwin)
		ps='ps auxwww'
		;;
	SunOS)
		ps='/usr/ucb/ps auxwww'
		;;
	*)
		ps='ps -ef'
		;;
esac

get_pid()
	{
	$ps | grep "$cmd" | grep -v grep | awk ' { print $2 } '
	}

case $1 in
	start)
		pid=`get_pid`
		if [ ! "$pid" = "" ] ; then
			echo "Already running SharedRep $loc $cluster as pid" $pid
		else
			$jcmd < /dev/null > /dev/null 2> /dev/null &
			pid=`get_pid`
			wait=
			while [ "$pid" = "" ] && [ ! "$wait" = "xxxxx" ] ; do
				sleep 1
				pid=`get_pid`
				wait=x$wait
			done
			if [ ! "$pid" = "" ] ; then
				echo "Started running SharedRep $loc $cluster as pid" $pid
			else
				echo "Can't run SharedRep $loc $cluster"
				exit 2
			fi
		fi
		;;
	stop|clean)
		pid=`get_pid`
		if [ ! "$pid" = "" ] ; then
			kill $pid
			pid=`get_pid`
			wait=
			while [ ! "$pid" = "" ] && [ ! "$wait" = "xxxxx" ] ; do
				sleep 1
				pid=`get_pid`
				wait=x$wait
			done
			if [ "$pid" = "" ] ; then
				echo "Stopped running SharedRep $loc $cluster"
			else
				echo "Couldn't stop running SharedRep $loc $cluster, pid" $pid
				exit 3
			fi
		else
			echo "Already not running SharedRep $loc $cluster"
		fi
		;;
	restart)
		$0 stop
		$0 start
		;;
	status)
		pid=`get_pid`
		if [ ! "$pid" = "" ] ; then
			echo "Running SharedRep $loc $cluster as pid" $pid
		else
			echo "Not running SharedRep $loc $cluster"
		fi
		;;
	monitor)
		pid=`get_pid`
		if [ ! "$pid" = "" ] ; then
			exit 110
		else
			exit 100
		fi
		;;
	*)
		echo "usage: $0 ( start | stop | clean | restart | status | monitor )"
		;;
esac

exit 0
